@@ -24,6 +24,11 @@
 
 #include "config.h"
 
+/* hack - CONFIG_SND_HDA_INPUT_JACK can be wrongly set for older kernels */
+#ifndef CONFIG_SND_JACK
+#undef CONFIG_SND_HDA_INPUT_JACK
+#endif
+
 /* number of supported soundcards */
 #ifdef CONFIG_SND_DYNAMIC_MINORS
 #define SNDRV_CARDS 32
@@ -1783,10 +1788,11 @@ typedef _Bool bool;
 #endif
 
 /* memdup_user() wrapper */
-#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 30)
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 30) || \
+	defined(CONFIG_SND_DEBUG_MEMORY)
 #include <linux/err.h>
 #include <asm/uaccess.h>
-static inline void *memdup_user(void __user *src, size_t len)
+static inline void *snd_memdup_user(const void __user *src, size_t len)
 {
 	void *p = kmalloc(len, GFP_KERNEL);
 	if (!p)
@@ -1797,6 +1803,17 @@ static inline void *memdup_user(void __user *src, size_t len)
 	}
 	return p;
 }
+#define memdup_user snd_memdup_user
 #endif
 
+/* PCI_VEDEVICE() */
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 20)
+#include <linux/pci.h>
+#ifndef PCI_VDEVICE
+#define PCI_VDEVICE(vendor, device)		\
+	PCI_VENDOR_ID_##vendor, (device),	\
+	PCI_ANY_ID, PCI_ANY_ID, 0, 0
+#endif
+#endif /* < 2.6.20 */
+
 #endif /* __SOUND_LOCAL_DRIVER_H */
